"""

"""


import traceback as tb

from os.path import exists
import re


class OMP:
    def __init__(self):
        self._tree = None
        self._list = None
        return

    def loads(self, text, path=None):
        node_list = []
        lines = text.split("\n")
        min_lvl = 1
        cur_lvl = min_lvl
        indent_rx = re.compile(r"^(\*+)\s(.+)")
        level = None

        for idx, l in enumerate(lines):
            l = l.strip()
            match = indent_rx.match(l)
            if not match:
                continue
            level = match.group(1).count("*")
            head = match.group(2)
            node_list.append(
                {
                    "level": level,
                    "head": head,
                    "body": [],
                    "lineno": idx + 1,
                    "col_offset": level,
                }
            )
        self._list = node_list
        node_tree = self._make_tree(node_list, min_lvl)
        if path:
            node_tree = self.extract_path(node_tree, path)
        self._tree = node_tree
        return node_tree

    def _make_tree(self, node_list, level=1):
        cur_lvl = level
        tree = []
        sub_tree = []

        for node in node_list:

            if node["level"] == cur_lvl:

                if sub_tree:
                    # recurse to connect deeper nodes
                    tree[-1]["body"] = self._make_tree(sub_tree, cur_lvl + 1)
                    sub_tree = []
                tree.append(node)

            elif node["level"] > cur_lvl:
                # build sublist
                sub_tree.append(node)

            else:
                # TODO: handle level-less text nodes
                continue

        if sub_tree:
            # end of file with no next parent-sibling
            tree[-1]["body"] = self._make_tree(sub_tree, cur_lvl + 1)
        return tree

    def load(self, path):
        if not exists(path):
            raise OSError(f"{path} not found")
        retval = None

        with open(path) as fo:
            text = fo.read()
            self._text = text
            retval = self.loads(text)
        self._tree = retval
        return retval

    def dumps(self):
        return

    def dump(self):
        return

    def apply(self, func, *args, **kwds):
        if not callable(func):
            raise ValueError("Applicator must be a callable")
        return func(self, *args, **kwds)

    @property
    def tree(self):
        return self._tree

    def extract_path(self, nt, path):
        st = nt  # sub tree

        while path:
            sp = path.pop(0)  # sub path
            sp = int(sp) if str(sp).isdigit() else re.compile(sp)

            if isinstance(sp, int):
                if 1 > sp >= len(st):
                    raise ValueError(
                        f"Length of subtree is {len(st)} but got associated subpath of {sp}"
                    )
                st = st[sp - 1]

            else:
                for node in st:
                    if sp.match(node.get("head", "")):
                        st = node.get("body")
                        break
        if not isinstance(st, list):
            st = [st]
        return st



"""
A resolver based on the Mycroft Adapt intent parser.
"""


import re
from time import time
from ast import Str, Expr, parse

from adapt.intent import IntentBuilder
from adapt.engine import IntentDeterminationEngine

from neulang.base import BaseResolver


class AdaptResolver(BaseResolver):
    def __init__(self):
        self._pifs = {}
        super().__init__()
        self._ad_test = re.compile(r"^add_adapt_handler(\s.+)?")
        self._eng = IntentDeterminationEngine()
        return

    def resolve(self, o_node):
        if not self.is_adapt(o_node):
            return False
        a_node = None
        head = o_node.get("head")
        h_match = self._ad_test.match(head)

        if h_match:
            a_node = self.add_handler(o_node)
            assert a_node is None, f"None expected but got {type(a_node).__name__}"

        else:
            a_node = self.nl_call(o_node)
        return a_node

    def add_handler(self, data):
        engine = self._eng
        ignores = ["intent_parts", "body"]
        data = self._node_to_list(data, ignore_first=ignores)
        if not len(data) in [1, 2]:
            raise ValueError(f"adapt handler takes 1 or 2 fields")
        matches = data[0].get("intent_parts")
        assert all(
            [isinstance(m, Str) for m in matches]
        ), f"Expected all strings for intent parts"
        itt_parts = [m.s for m in matches]
        th = f"ad_{abs(hash(time()))}"

        for idx, ip in enumerate(itt_parts[:]):
            # hack if no named groups in rx
            ip_fix = ip

            while re.search(r"<<\w+>>", ip_fix):
                ip_fix = ip_fix.replace("<<", "(?P<").replace(">>", ">.+)")
            ip = ip_fix
            if not ("req_" in ip or "opt_" in ip):
                ip = f"(?P<req_{th}__{idx}>{ip})"
            itt_parts[idx] = ip
        topic = "neulang.resolvers.adapt.intent_parts_lists"
        self._pubsub.pub(topic, itt_parts)
        ib = IntentBuilder(f"{th}")
        ipn_rx = re.compile(r"P<\w+>.")

        for part in itt_parts:
            intermed = [
                ipn[2:-2]
                if not ipn[2:-2].isidentifier()
                else ib.optionally(ipn[2:-2].partition("_")[-1])
                if "opt_" in ipn
                else ib.require(ipn[2:-2].partition("_")[-1])
                if "req_" in ipn
                else None
                for ipn in ipn_rx.findall(part)
            ]
            if None in intermed:
                cnt = intermed.count(None)
                raise ValueError(
                    f'found {cnt} named group{"s" if cnt > 1 else ""} missing a "req_" or "opt" in "{part}"'
                )
            bad = [i for i in intermed if isinstance(i, str)]
            if any(bad):
                n = "are" if len(bad) > 1 else "is"
                raise NameError(
                    "all intent part names must be valid identifiers; {} {} not".format(
                        '", "'.join(bad), n
                    )
                )
            part = part.replace("opt_", "").replace("req_", "")
            engine.register_regex_entity(part)
        itt = ib.build()
        engine.register_intent_parser(itt)
        if len(data) == 1:
            return None
        body = data[1].get("body", [])
        body = [Expr(a_n) if not type(a_n) in self._stmt_types else a_n for a_n in body]
        pif = body or []
        self._pifs[th] = pif
        return None

    def nl_call(self, data):
        """Convert a natural language 'call' into a list of nodes."""
        head = data.get("head")
        body = data.get("body", [])
        itt = context = p_body = None
        engine = self._eng
        anl = []

        if body:
            # process body
            p_body = self._node_to_list(data)
        intents = [i for i in engine.determine_intent(head)]
        if not intents:
            return False
        itt = intents[0]
        bit = itt.get("intent_type")
        pif = self._pifs.get(bit)
        if not pif:
            return False
        assert isinstance(pif, list), f"Unable to resolve `{head}`"
        del itt["intent_type"]
        del itt["confidence"]
        itta = [
            parse(f'{k}="{v}"').body[0]
            for k, v in itt.items()
            if not k.startswith(f"{bit}__")
        ]
        anl.extend(itta)
        th = f"nl_{abs(hash(time()))}"

        if p_body:
            im = []
            left = [
                im.append(nd)
                if hasattr(nd, "lineno")
                else im.extend(nd)
                if isinstance(nd, list) and all([hasattr(n, "lineno") for n in nd])
                else nd
                for nd in p_body
                if nd
            ]
            assert not any(left), "Some items not properly processed"
            anl.extend(im)
            self._pifs[bit.replace("ad", "nl")] = im

        else:
            p_body = self._pifs.get(bit.replace("ad", "nl"), [])
            anl.extend(p_body)
        anl.extend(pif)
        return anl

    def is_adapt(self, node):
        # Just a stub until a way to detect
        return True



"""
Base resolver and registry.
"""


import traceback as tb

from ast import *
import sys
import re
from shlex import shlex


class ResolverRegistry(type):
    _registry = {}
    _instances = {}

    def __new__(cls, name, bases, attrs):
        # register and instanciate class as soon as it's defined
        new_cls = type.__new__(cls, name, bases, attrs)
        cls._registry[new_cls.__name__] = new_cls
        new_cls()
        return new_cls

    def __call__(cls, *args, **kwds):
        # make registered classes singletons
        if cls not in cls._instances:
            cls._instances[cls] = super(ResolverRegistry, cls).__call__(*args, **kwds)
        return cls._instances[cls]

    @classmethod
    def get_resolvers(cls):
        return cls._instances

    @classmethod
    def add_resolver(cls, resolver):
        # allow usage as non-meta
        cls._instances[resolver.__class__.__name__] = resolver
        return True


class BaseResolver:
    # __metaclass__ = ResolverRegistry

    def __init__(self,):
        ResolverRegistry.add_resolver(self)
        self._stmt_types = [
            Assign,
            Import,
            ImportFrom,
            If,
            For,
            While,
            FunctionDef,
            ClassDef,
            Delete,
            Pass,
            Break,
            Continue,
            Return,
            Yield,
            YieldFrom,
            Global,
            Nonlocal,
            Await,
            Raise,
            Assert,
            Try,
            ExceptHandler,
            With,
            AsyncFor,
            AsyncWith,
        ]
        self._resolvers = self.get_resolvers()
        self._top_level = 1
        return

    def get_resolvers(self):
        return ResolverRegistry.get_resolvers()

    def _resolve(self, node):
        """Resolve a node by seeking the correct resolver"""
        # TODO: resolver priority list or sort method
        a_node = None
        anl = []
        resolvers = self._resolvers
        r_names = [
            "ASTIR",
            "AdaptResolver",
            # "RegexResolver",
            "PyExprResolver",
        ]
        resolved = False
        is_single = True  # flag to de-list single-node results

        for rsvr in r_names:
            # TODO: make custom return object to handle special cases
            if rsvr not in resolvers:
                raise KeyError(f'"{rsvr}" not found in loaded resolvers')
            a_node = resolvers[rsvr].resolve(node)
            if a_node is False:
                continue  # node unhandled flag

            if a_node is None:
                # nothing to add (but node already handled)
                resolved = True
                break

            if a_node:
                # make list for processing
                if not isinstance(a_node, list):
                    is_single = False
                    a_node = [a_node]
                assert not any(
                    [isinstance(e, list) for e in a_node]
                ), "Found nested list"
                a_node = [self.add_location(a_n, node) for a_n in a_node]

                if node.get("level") == self._top_level:

                    for a_n in a_node:
                        if (
                            not isinstance(a_n, Expr)
                            and not type(a_n) in self._stmt_types
                        ):
                            a_n = Expr(a_n)
                        a_n.lineno = node.get("lineno", 0)
                        a_n.col_offset = node.get("col_offset", 0)
                        self.pre_test(a_n, node.get("lineno", 0))
                        anl.append(a_n)
                resolved = True
            if not anl:
                anl = a_node
            if anl:
                break
        if not resolved:
            raise ValueError("Unresolved node")
        if not is_single:
            return anl[0]
        return anl

    def add_handler(self, data):
        pass

    def add_location(self, a_node, o_node):
        """Add locations from org to AST node"""
        lineno = o_node.get("lineno", None)
        col_offset = o_node.get("col_offset", None)
        if not (lineno or col_offset):
            return a_node

        def _fix(node, lineno, col_offset):
            if "lineno" in node._attributes:
                if not hasattr(node, "lineno"):
                    node.lineno = lineno
                else:
                    lineno = node.lineno
            if "col_offset" in node._attributes:
                if not hasattr(node, "col_offset"):
                    node.col_offset = col_offset
                else:
                    col_offset = node.col_offset
            for child in iter_child_nodes(node):
                _fix(child, lineno, col_offset)

        _fix(a_node, lineno or 1, col_offset or 0)
        return a_node

    def cast_air(self, value, mode="norm"):
        # cast to type via wrapper
        # TODO: prob delegate to PyExprResolver
        node = None

        if self.is_num(value):
            # number
            node = Num(n=float(value) if "." in value else int(value))

        elif value in ["True", "False", "None"]:
            node = NameConstant(
                True if value == "True" else False if value == False else None
            )

        elif self.is_name(value):
            node = Name(value, Load())

        elif self.is_string(value):
            # quoted string; maybe make fallthrough
            node = Str(s=value[1:-1])

        else:
            # unquoted string (or sth else)
            node = Str(value) if mode == "norm" else None
            # raise TypeError(f'Unable to detect a castable type for `{value}`')
        return node

    def _handle_inline_node(self, data, mode, ignore_first=False):
        # node with args in head
        data = self._behead(data)

        if mode == "split":
            data = data.split(" ")

        elif mode == "token":
            # tokenize to account for quotes
            data = self._tokenize(data)

        elif mode == "single":
            # make a list
            data = [data]

        else:
            raise ValueError(f"Unknown mode {mode}")
        anl = [self.cast_air(v) for v in data]
        return anl

    def _handle_full_node(self, data, ignore_heads=False):
        # args from node with body
        if not isinstance(data, list):
            raise ValueError(f"Expected a list, but got {type(data).__name__}")
        anl = []  # AST node list

        for o_node in data:
            head = o_node.get("head")
            a_node = (
                {head: self._handle_full_node(o_node.get("body"))}
                if isinstance(ignore_heads, list) and head in ignore_heads
                else self._resolve(o_node)
            )
            if (
                a_node is False
                or isinstance(a_node, dict)
                and a_node.get(o_node.get("head")) is False
            ):
                return False  # Wut!!! TODO: watch
            if a_node in [None, []]:
                continue
            anl.append(a_node) if not isinstance(a_node, list) else anl.extend(a_node)
        return anl

    def _handle_combo_node(self, data):
        # node with args in head and body
        return node

    def _behead(self, data):
        # remove prefix if it's there
        if not isinstance(data, str):
            raise ValueError(f"Expected a string but got {type(data).__name__}")
        if re.search(r"air_[A-Za-z_]+(\s.*)?", data):
            data = data.split(" ", 1)[1]
        return data

    def _node_type(self, node):
        nt = ""
        head = node.get("head")
        body = node.get("body")

        if not body:
            nt = "inline"

        elif not " " in head:
            nt = "full"

        elif " " in head and body:
            nt = "combo"

        else:
            nt = "unknown"
        return nt

    def _node_to_list(self, node, mode="token", ignore_first=False):
        # convert a node to a list, and any sub-nodes to AST
        data = None

        if self.is_inline(node):
            data = self._handle_inline_node(node.get("head"), mode)

        else:
            data = self._handle_full_node(node.get("body"), ignore_first)
        return data

    def _tokenize(self, text):
        # tokenize string into args
        if not isinstance(text, str):
            raise TypeError(f"Expected a str but got {type(text)}")
        tokens = []
        sl = shlex(text)

        while True:
            token = sl.read_token()
            if not token:
                break
            tokens.append(token)
        return tokens

    def is_inline(self, node):
        if not isinstance(node, dict):
            raise TypeError(f"Expected a dict but got {type(node).__name__}")
        return self._node_type(node) == "inline"

    def is_full(self, node):
        if not isinstance(node, dict):
            raise TypeError(f"Expected a dict but got {type(node).__name__}")
        return self._node_type(node) == "full"

    def is_combo(self, node):
        if not isinstance(node, dict):
            raise TypeError(f"Expected a dict but got {type(node).__name__}")
        return self._node_type(node) == "combo"

    def is_name(self, text):
        # check if something is a valid Python name
        if not isinstance(text, str):
            raise TypeError(f"Expected a str but got {type(text).__name__}")
        return not not re.search(r"^[A-Za-z_][A-Za-z0-9_]*$", text)

    def is_string(self, text):
        # check if a quoted string
        if not isinstance(text, str):
            raise TypeError(f"Expected a str but got {type(text).__name__}")
        return text[:1] in ['"', "'"] and text[:1] == text[-1:]

    def is_num(self, value):
        if not isinstance(value, str):
            raise TypeError(f"Expected a str but got {type(value).__name__}")

        try:
            float(value)
            return True

        except ValueError:
            return False

    def pre_test(self, node, line):

        try:
            nm = Module([node])
            nc = compile(nm, "<pre_test>", "exec")

        except Exception as e:
            print(f"Compile test failed at line #{line}")
            tb.print_exc()
            sys.exit(1)
        return node



"""

"""


import pytest

from neulang.callees import Callee


class TestCallee:

    @pytest.fixture
    def c(self):
        return Callee()

    def test_x_setv(self, c):
        c.x_setv('a', 3)
        assert c.ns['a'] == 3
        return

    def test_x_getv(self, c):
        c.x_setv('b', 29)
        assert c.x_getv('b') == 29
        return

    def test_c_delv(self, c):
        c.x_setv('q', 13)
        c.x_delv('q')
        assert 'q' not in c.ns
        return

    @pytest.mark.parametrize('test, t_body, f_body, ctx, e_retv', [
        (lambda **ctx: True, lambda **ctx: 'test 1', None, {}, 'test 1'),
        (lambda **ctx: False, lambda **ctx: 'test 2 true', lambda **ctx: 'test 2 false', {}, 'test 2 false'),
    ])
    def test_x_if(self, c, test, t_body, f_body, ctx, e_retv):
        # TODO: test expr test and other possible cases
        retv = c.x_if(test, t_body, f_body, **ctx)
        assert retv == e_retv
        return

    @pytest.mark.parametrize('target, iter_, body, orelse, ctx, e_retv', [
        ('idx', range(0, 5), lambda i, **ctx: f'{i} potato', lambda **ctx: 'no potato', {}, 'no potato'),
    ])
    def test_x_for(self, c, target, iter_, body, orelse, ctx, e_retv):
        # TODO: test other possible cases
        retv = c.x_for(target, iter_, body, orelse, **ctx)
        assert retv == e_retv
        return

    @pytest.mark.parametrize('test, body, orelse, ctx, e_retv', [
        ('5 == 3 + 1', lambda ctx: True, None, {}, None),
    ])
    def test_x_while(self, c, test, body, orelse, ctx, e_retv):
        # TODO: more input combos
        retv = c.x_while(test, body, orelse, **ctx)
        assert retv == e_retv
        return

    def test_x_do(self, c):
        from functools import partial
        ctx = {}

        def add(val, **ctx):
            return val + ctx.get('return', 0)
        one = partial(add, 1)
        two = partial(add, 2)
        three = partial(add, 3)
        retv = c.x_do(one, two, three, **ctx)
        assert retv == 6
        return

    def test_x_import(self, c):
        mods = c.x_import('os', 'sys')
        assert mods and isinstance(mods, tuple)
        mod = c.x_import('.callees', packages='neulang')
        assert mod
        return

    def test_x_with(self, c):
        def enter(self):
            return 'entered'
        CM = c.x_class('CM', {'__enter__': enter, '__exit__': lambda s, *a: None})
        retv = c.x_with([CM, 'rv'], lambda rv, **ctx: rv)
        assert retv == 'entered'
        return

    def test_x_class(self, c):
        meth = lambda self: 3
        Cls = c.x_class('MyException', {'meth': meth}, tuple([Exception]))
        cls = Cls()
        assert isinstance(cls, Exception)
        assert cls.meth() == 3
        return

    @pytest.mark.parametrize('body, hans, ore, fin, ctx, e_retv', [
        (lambda c, **ctx: c.x_raise(SyntaxError, 'test'), [[SyntaxError, lambda exc, **ctx: 'syntax error raised']], None, None, {}, 'syntax error raised')
    ])
    def test_x_try(self, c, body, hans, ore, fin, ctx, e_retv):
        # TODO: test other possible cases
        from functools import partial
        retv = c.x_try(partial(body, c), *hans, orelse=ore, finalbody=fin, **ctx)
        assert retv == e_retv
        return

    def test_x_raise(self, c):
        msg = 'test successful'
        try:
            c.x_raise(ValueError, msg)

        except SyntaxError as se:
            assert se.args[0] == 'something broke'

        except ValueError as ve:
            assert ve.args[0] == msg

        except Exception as e:
            assert e.args[0] == 'more breakage'
        return

    def test_to_dict(self, c):
        cd = c.to_dict()
        assert cd and isinstance(cd, dict) and all([k.startswith('x_') for k in cd.keys()]) and all([callable(v) for v in cd.values()])
        return



"""

"""


import pytest

from neulang.pubsub import PubSub


class TestPubSub:

    @pytest.fixture
    def ps(self):
        return PubSub()

    def test_pub(self, ps):
        t = 'namespace.default'
        c = 'my content'
        ps.pub(t, c)
        hub = ps._hub
        assert t in hub and hub[t].get('content_list') == [c]
        return

    def test_sub(self, ps, mocker):
        t = 'namespace.default'
        self.test_pub(ps)
        c = mocker.Mock()
        nsl = ps.sub()
        assert nsl == [t]
        ps.sub(topic=t, callback=c)
        c.assert_called_with('my content')
        return



#! /usr/bin/env python3
"""

"""


import traceback as tb

from ast import *
from os.path import exists as op_exists, join as op_join, isfile as op_isfile
import re
import sys

try:
    import astor

except:
    pass

from neulang.omp import OMP
from neulang.resolvers import Resolver
from neulang.pubsub import PubSub


if not sys.version_info[:2] == (3, 7):
    msg = "Neulang was created and tested with Python 3.7, but detected {}. YMMV.".format(
        ".".join([str(n) for n in sys.version_info[:3]])
    )
    import warnings

    warnings.warn(msg, RuntimeWarning)


class AttribProxy:
    """Generate access via attributes"""

    # TODO: move to utils

    def __init__(self, data):
        self._data = data
        # self._func = func
        # self._args = args
        # self._kwds = kwds
        return

    def __getattr__(self, attr):
        data = self._data
        if isinstance(data, dict):
            return data[attr]
        if callable(data):
            return data(attr)
        raise AttributeError("Unable to process {attr}")


class Neulang:
    def __init__(self, *args, **kwds):
        ns = kwds.get("namespace", None)
        self._ns = ns = ns if isinstance(ns, dict) else {}
        self._ps = ps = PubSub()
        self._resolver = Resolver(parent=self, pubsub=ps)
        self._ast = None
        self._src = None
        self._latest = None
        self._org_path = None
        ns["include"] = self.load_module
        ns["__PUBSUB__"] = ps
        return

    def load_astir(self, af=""):
        ns = {}
        if not exists(af):
            return
        am = self.load(af)
        am = compile(am, af, "exec")
        exec(am, ns)
        ap = AttribProxy(ns)
        return ap

    def load(self, path):
        if not op_exists(path):
            raise OSError(f"{path} not found")
        success = None

        mod = open(path).read()

        try:
            success = self.loads(mod, sf=path)

        except AssertionError as e:
            tb.print_exc()
            print("\n\nPlease report this error to the developer")
            return False

        except Exception as e:
            tb.print_exc()
            return False
        return success

    def loads(self, text, sf="<module>"):
        """Convert org text to AST object"""
        ompo = OMP()
        ompo.loads(text, self._org_path)

        try:
            ast = self._resolver.resolve(ompo.tree)

        except Exception as e:
            raise
        assert isinstance(
            ast, AST
        ), f'expected an AST object but got "{ast.__class__.__name__}" object instead'
        self._ast = ast
        self._src = sf
        self._latest = False
        return True

    def eval(self, mode="exec"):
        if self._latest:
            return
        rv = None

        if mode == "exec":
            exec(compile(self._ast, self._src, "exec"), self._ns)

        elif mode == "eval":
            ast = self._ast
            if isinstance(ast, Module):
                ast = ast.body[0]
                ast = Expression(ast.value if isinstance(ast, Expr) else ast)
            rv = eval(compile(ast, self._src, "eval"), self._ns)
            self._ns["_"] = rv
        self._latest = True
        return rv

    def reset(self):
        self._ast = self._ns = None
        self._latest = True
        return

    def org_path(self, op):
        if not op or not isinstance(op, list):
            return False
        self._org_path = op
        return True

    @property
    def ns(self):
        return self._ns

    def update_namespace(self, *r_list, **a_dict):
        if (r_list and not all([i.isidentifier() for i in r_list])) or (
            a_dict and not all([k.isidentifier() for k in a_dict.keys()])
        ):
            raise NameError("All items and keys must be valid names")
        ns = self._ns

        for name in r_list:
            if name in ns:
                del ns[name]

        for k, v in a_dict.items():
            ns[k] = v
        return True

    def to_py(self):
        if "astor" not in globals():
            raise ImportError(
                'Cannot convert to py as astor module not found. Run "pip install astor" and try again.'
            )
        ast = self._ast
        if not ast:
            raise ValueError("No source code loaded.")
        py = astor.to_source(ast)
        return py

    def load_module(self, name, package=None):
        res = False
        f_name = "{}.neu".format(op_join(".", name))

        if not "." in name and op_exists(f_name):
            res = self.load(f_name)
            return res
        n_parts = name.split(".")
        s_paths = sys.path  # TODO: get path from ns
        assert all([isinstance(p, str) for p in s_paths]), "All paths must be strings"
        if not "" in s_paths:
            s_paths.insert(0, "")
        checked = []
        m_paths = []

        for path in s_paths:
            if not path:
                path = "."
            if path in checked:
                continue
            checked.append(path)
            if op_exists(op_join(path, n_parts[0], "__init__.neu")):
                success = self._name_to_paths(path, n_parts[:], m_paths)
                if success and m_paths:
                    m_paths = m_paths[::-1]
                    break
        else:
            raise ModuleNotFoundError(f'Cannot resolve "{name}"')

        for mp in m_paths:
            try:
                res = self.load(mp)
                self.eval()

            except Exception as e:
                raise ImportError(f'unable to process "{mp}"; {e.args[0]}')
        return res

    def _name_to_paths(self, path, n_parts, m_paths, p_parts=None):
        if not p_parts:
            p_parts = []
        np = n_parts.pop(0)
        p_parts.append(np)
        nfs = [
            op_join(path, *p_parts, "__init__.neu"),
            "{}.neu".format(op_join(path, *p_parts)),
        ]
        if not (op_isfile(nfs[0]) or op_isfile(nfs[1])):
            return False

        if not n_parts:
            if op_exists(nfs[1]):
                m_paths.append(nfs[1])
            if op_exists(nfs[0]):
                m_paths.append(nfs[0])
            return True
        success = self._name_to_paths(path, n_parts, m_paths, p_parts)
        if success:
            assert op_exists(nfs[0]), f'"{nfs[0]}" not found'
            m_paths.append(nfs[0])
        return success

    def sub(self, *args, **kwds):
        return self._ps.sub(*args, **kwds)



#! /usr/bin/env python3

"""
Neulang: Executable org-formatted pseudocode embedded in Python.
Usage:
    neu [options] [<file>|(--command=<cmd>)]  [(--org-path=<op>)] [-m<modules>] [--to-py]

-v, --version   Print version info.
-i              Open interpreter after executing command.
-o, --org-path=<op>
                Internal traversal path as list of regexes or indices, using sed format (works with <file>).
-m, --modules=<mods>
                List of modules to import, separated by ":".
--command=<cmd>
                Execute a command.
--to-py         Convert code to Python source.
"""


import traceback as tb

import sys
import json
from os.path import exists, join, dirname

try:
    from docopt import docopt

except:
    # likely just installing
    pass

from neulang import Neulang, Callee


__all__ = ("__meta__", "__version__", "run")
meta_file = join(dirname(__file__), "meta.json")
if not exists(meta_file):
    open(meta_file, "w").write("{}")
__meta__ = json.load(open(meta_file))
__author__ = __meta__.get("author", "skeledrew")
__version__ = __meta__.get("version", "0.1.0")


def _run():
    # catch assert errors

    try:
        run()

    except AssertionError as ae:
        tb.print_exc()
        sys.exit(1)
    return


def run():
    opt = docopt(__doc__)
    src_file = opt.get("<file>") or ""
    q_cmd = opt.get("--command") or None
    org_path = opt.get("--org-path") or ""
    org_path = (
        org_path.split(org_path[1])[1:]
        if org_path.startswith("s") and len(org_path) > 2
        else None
    )
    mods = opt.get("--modules") or []
    gen_py = opt.get("--to-py") or None
    if mods:
        mods = mods.split(":")
    neulang = Neulang()
    callee = Callee(namespace=neulang.ns)
    neulang.update_namespace(**callee.to_dict())

    for mod in mods:
        if not exists(mod):
            raise OSError(f'"{mod}" not found.')
        neulang.load(mod)
        neulang.eval()

    if q_cmd:
        if not q_cmd.startswith("* "):
            q_cmd = f"* {q_cmd}"
        neulang.loads(q_cmd)
        neulang.eval()

    if src_file:
        neulang.org_path(org_path)
        neulang.load(src_file)

        if not gen_py:
            neulang.eval()

        else:
            try:
                code = neulang.to_py()
                print(code)

            except Exception as e:
                print(f"{repr(e)}")

    if not (src_file or mods) or opt.get("-i"):
        # enter REPL
        level = 1
        accum = []
        i_char = "\\"
        d_char = "/"
        q_words = ["*quit*", "*exit*"]
        print(f"Neulang {__version__}")
        print(f'Use "{i_char}" or "{d_char}" to indent or deindent nodes.')
        print(f"Evaluate code by entering an empty node after.")
        print(f'Type "{q_words[0]}" or "{q_words[1]}" to quit.\n')

        while True:
            # basic interpreter
            rv = ""
            stars = "*" * level
            prompt = f"{stars} "

            try:
                inp = input(prompt).strip()

            except KeyboardInterrupt:
                print("\nKeyboardInterrupt")
                continue

            except EOFError:
                print("")
                break

            if inp in q_words:
                break

            if accum and not inp:

                try:
                    neulang.loads("\n".join(accum))

                    try:
                        rv = neulang.eval("eval")
                        if rv is None:
                            rv = ""

                    except (SyntaxError, TypeError):
                        # NB: catching TypeError might mask pertinent errors
                        neulang.eval()

                    else:
                        print(rv)

                except KeyboardInterrupt:
                    continue

                except Exception as e:
                    tb.print_exc()

                finally:
                    level = 1
                    accum = []
                continue

            if inp is i_char or inp.endswith(i_char):
                inp = inp.rstrip(i_char)
                level += 1

            elif inp is d_char or inp.endswith(d_char):
                inp = inp.rstrip(d_char)
                level -= 1
                if level < 1:
                    level = 1
            if inp:
                accum.append(f"{prompt}{inp}")

    elif opt.get("--version", None):
        print(f"Neulang {__version__}")
    return


if __name__ == "__main__":
    run()



"""

"""


import pytest

from neulang.omp import OMP
from neulang.resolvers import PyExprResolver

from neulang.adapt_resolver import AdaptResolver


class TestAdaptResolver:

    @pytest.fixture
    def otree(self):
        def otree(text):
            ompo = OMP()
            ompo.loads(text, None)
            return ompo.tree
        return otree

    @pytest.fixture
    def ar(self):
        return AdaptResolver()

    @pytest.mark.xfail(reason='TBD')
    def test_resolve(self):
        assert 0
        return

    @pytest.mark.xfail(reason='TBD')
    def test_add_handler(self, otree, ar, mocker, monkeypatch):
        t_rx = 'get (?P<req_var>\w+)'
        d = f'* add_adapt_handler\n** intent_parts\n*** {t_rx}'
        rre_mock = mocker.Mock(side_effect=ar._eng.register_regex_entity)
        monkeypatch.setattr(ar._eng, 'register_regex_entity', rre_mock)

        try:
            res = ar.add_handler(otree(d)[0])

        except Exception as e:
            res = e
        assert res is None
        rre_mock.assert_called_with(t_rx)
        return

    @pytest.mark.xfail(reason='TBD')
    def test_nl_call(self):
        assert 0
        return



"""
Tests for core.py
"""


import pytest

from os.path import exists

import neulang.core as core
from neulang.callees import Callee


def neu_samples():
    # TODO: find a way to always locate tests.neu
    tests = """
    * air_say
** air_joined_str
*** air_mul
**** air_div
***** 48
***** 94
**** 100
*** '% of AST nodes implemented'

* print("This is a regular Python expression...")

* air_say '... and this is inline ASTIR (Abstract Syntax Tree Intermediate Representation)...'

* air_setv
** this
** "... which can also be expressed hierarchially"
* air_say this

* air_say
** ["a", 'Python', "list"]
** is also a regular Python expression,
** "but under ASTIR"

* air_say
** 'Now for a lil math: 7+13+9='
** air_add
*** 7
*** 13
*** air_add 6 3

* print(f'15+21={15 + 21} also works of course')

* air_say
** 'Testing random combined math:'
** air_mul
*** 5
*** 31
*** air_div 10 3
*** 7.9
*** air_sub 99.3 100
** with some text after it

* air_say
** Testing unary 'not'; should yield True:
** air_not 0
** and False:
** air_not
*** 13

* air_say
** 'And' table results:
** air_and
*** 0
*** 0
** air_and
*** 0
*** 1
** air_and
*** 1
*** 0
** air_and 1 1

* air_say
** 'Or' table results:
** air_or
*** 0
*** 0
** air_or 0 1
** air_or
*** 1
*** 0
** air_or 1 1

* air_say
** Test comparison 5 == 4+1 == 3+2:
** air_eq
*** 5
*** air_add 4 1
*** air_add
**** 3
**** 2
*** 
** 

* air_call print 'Now we can call functions!'

* air_nop
** Everything in this node is ignored
** air_call
*** air_say
*** This call syntax looks a bit weird...
*** But it'll make sense in time, I hope
** air_call
*** air_say
*** This string has embedded "double quotes" and 'single quotes' in it, and even "a 'nested example'".
** air_while
*** 
** air_return None

* air_import
** os
* air_import random
* air_say 'Imports:' os random

* air_if
** air_eq
*** 5
*** 6
** air_say "They're equal!"
** air_say "They aren't equal..."

* air_for
** idx
** air_call range 1 5
** air_say idx "potato"

* air_def
** hello
** args
*** name
** body
*** air_say 'Hi there' name '!'
** decorators
** returns
* hello('Drew')

* air_class
** my_life
** body
*** air_setv fav_num 13
*** air_setv
**** now_playing
**** "Chat Wars"
** 
* print(f"Favorite number is {my_life.fav_num}")
* air_say
** Current obsession:
** air_call
*** getattr
*** my_life
*** 'now_playing'
* air_delete my_life

* air_assert True 'Some random message that never shows'

* air_try
** body
*** air_say 'Testing exceptions'
*** air_raise
**** air_call
***** Exception
***** Some random reason
** handlers
*** air_except_handler
**** Exception
**** e
**** body
***** air_say 
****** "Raising for:"
****** air_call str e

* air_say
** air_dict
*** 145
**** 'one quatro go'
*** 'a string'
**** 'says blah'
*** True
**** 13

* air_def
** arg_ts
** args
*** a
*** air_kwd
**** e
**** 27
*** air_starred
**** b
*** air_kwd
**** c
**** None
*** air_kwd
**** d
** body
*** air_say a b c d e
* air_call
** arg_ts
** 29
** air_starred
*** air_list
**** 'a disassociated list' 
**** 'of strings'
** air_kwd
*** c
*** True
** air_kwd
*** air_dict
**** 'x'
***** 'ekans'
**** 'y'
***** 'Yugioh'
**** 'z'
***** 'Zabuza'

* air_say
** 'Testing'
** air_list 'list' 'of' 'elements'

* air_say
** 'Result of slicing a list: '
** air_slice
*** air_list 1 2 3
*** -1

* air_with
** items
*** fo
**** open('file.txt', 'w')
** body
*** fo.write('Some random text')

* air_say 'Goodbye ASTIR'

* print("=============================")

* say Hello natural language programming
** add_adapt_handler
*** intent_parts
**** 'say (?P<req_text>.+)'
*** body
**** air_say text
* say The above sub definition is visible beyond!
"""
    src = open('tests/tests.neu').read() if exists('tests/tests.neu') else tests
    samps = [e for e in src.split('\n\n') if e.startswith('* ')] if '\n\n' in src else list(map(lambda l: f'* {l}', '\n'.join(list(filter(lambda l: l[:1] == '*', src.split('\n')))).split('\n* ')))
    return samps

@pytest.fixture(scope="module")
def Neu():
    neu = core.Neulang
    return neu

@pytest.fixture(scope="function")
def neu(Neu):
    neu = Neu()
    c = Callee()
    neu.update_namespace(**c.to_dict())
    return neu


class TestNeulang:

    @pytest.mark.parametrize("expr,retv", [("* 1 + 5", 6)])
    def test_eval_py_expr(self, neu, expr, retv):
        neu.loads(expr)
        assert neu.eval(mode="eval") == retv
        return

    @pytest.mark.parametrize("expr, fun, mode", [
        ("* air_setv\n** a\n** 3", lambda ns: ns['a'] == 3, 'exec'),
        ("* air_slice\n** air_list 1 2 3\n** -1", lambda ns: ns['_'] == 3, 'eval'),
        ('* air_call\n** print\n** Weird call syntax', lambda ns: True, 'exec'),
        ("""* air_setv
** string
** 'This is a string'
* add_adapt_handler
** intent_parts
*** 'replace (?P<req_txt>.+) with (?P<req_rep>.+) in (?P<req_var>\w+)'
** body
*** air_setv
**** globals()[var]
**** globals()[var].replace(txt, rep)
* replace string with modified string in string""", lambda ns: ns['string'] == 'This is a modified string', 'exec'),
        ("""
* x_setv('b', 7)
* air_setv
** q
** x_getv('b') * 12""", lambda ns: ns['q'] == 84, 'exec'),
        ("""
* find average of 3 and 5
** add_adapt_handler
*** intent_parts
**** 'find average of (?P<req_num1>\d+) and (?P<req_num2>\d+)'
*** body
**** x_setv('num1', num1)
** get and sum the numbers
*** add_adapt_handler
**** intent_parts
***** 'get and sum the numbers'
**** body
#***** print(x_setv('num1'), num1)
***** x_setv('num_sum', int(num1) + int(num2))
** divide by 2
*** add_adapt_handler
**** intent_parts
***** 'divide by (?P<req_amount>\d+)'
**** body
***** x_setv('num_ave', x_getv('num_sum') / int(amount))
** output result
*** add_adapt_handler
**** intent_parts
***** 'output result'
**** body
***** air_setv
****** num_ave
****** x_getv("num_ave")
***** print(f'Average of {num1} and {num2} is {num_ave}')""", lambda ns: ns['num_ave'] == 4, 'exec')
    ])
    def test_eval_astir(self, neu, mocker, monkeypatch, expr, fun, mode):
        neu.loads(expr)
        neu.eval(mode=mode)
        assert fun(neu._ns)
        return

    @pytest.mark.parametrize("samp", [s for s in neu_samples()])
    def test_samples(self, neu, samp, capsys):
        # Run the samples in tests.neu
        # NB: doesn't really validate anything it seems
        neu.loads(samp)
        neu.eval()
        captured = capsys.readouterr()
        assert captured and not 'Error' in captured
        captured = None
        return

    def test_update_namespace(self, neu):
        ad = {'a': 23, 'z': False}
        rl = ['z']
        neu.update_namespace(**ad)
        assert neu.ns['a'] == 23 and neu.ns['z'] is False
        neu.update_namespace(*rl)
        assert 'z' not in neu.ns
        return

    def test_to_py(self, neu):
        n_s = neu_samples()
        n_code = n_s[0]
        neu.loads(n_code)
        py = neu.to_py()
        assert 'print' in py and 'AST nodes' in py
        return

    @pytest.mark.parametrize('name, pkg, code, e_path, e_res_t', [
        ('script', None, '* neu code', './script.neu', str),
        ('foo.bar', None, '* neu code', './foo/bar/__init__.neu', str),
        ('foo.bar.bad', None, 'bad\nneu\ncode', './foo/bar/bad.neu', ImportError),
        ('fer.ben', None, '* invalid package/module format', None, ModuleNotFoundError),
        ('foo.bar.baz', None, '* neu code', './foo/bar/baz.neu', str)
    ])
    def test_load_module(self, neu, mocker, monkeypatch, name, pkg, code, e_path, e_res_t):
        m_paths = [
            './foo',
            './foo/__init__.neu',
            './foo/bar',
            './foo/bar/__init__.neu',
            './foo/bar/bad.neu',
            './fer',
            './fer/ben.neu',
            './script.neu',
            './foo/bar/baz.neu',
        ]

        def ope(p):
            return True if p in m_paths else False
        ope_mock = mocker.Mock(side_effect=ope)
        opi_mock = mocker.Mock(side_effect=lambda p: True if p.endswith('.neu') else False)

        def n_eval(*a, **k):
            neu.loads(code)
            return True #neu.eval(*a, **k)  # RecursionError
        l_mock = mocker.Mock(side_effect=lambda m, sf='blah': code)
        monkeypatch.setattr(neu, 'load', l_mock)
        monkeypatch.setattr(core, 'op_exists', ope_mock)
        monkeypatch.setattr(core, 'op_isfile', opi_mock)
        monkeypatch.setattr(neu, 'eval', n_eval)

        try:
            res = neu.load_module(name, pkg)
            l_mock.assert_called_with(e_path)

        except Exception as e:
            res = e
        assert isinstance(res, e_res_t)
        return



#! /usr/bin/env python3


"""

"""


import traceback as tb

from ast import *
import sys
import re

from .base import BaseResolver


class ASTIR(BaseResolver):
    def __init__(self):
        super().__init__()
        self._binop_map = {"add": Add(), "sub": Sub(), "mul": Mult(), "div": Div()}
        self._unaryop_map = {"not": Not()}
        self._boolop_map = {"and": And(), "or": Or()}
        self._compare_map = {
            "eq": Eq(),
            "not_eq": NotEq(),
            "lt": Lt(),
            "lt_e": LtE(),
            "gt": Gt(),
            "gt_e": GtE(),
        }
        self._control_flow_map = {
            "if": If,
            "for": For,
            "while": While,
            "try": Try,
            "except": ExceptHandler,
            "async_for": AsyncFor,
        }
        return

    def resolve(self, o_node):
        """Parse an OMP node into AST"""
        if not self.is_astir(o_node):
            return False
        a_node = None
        head = o_node["head"]
        body = o_node.get("body")
        opr = head.split(" ", 1)[0]
        level = o_node.get("level", 0)

        try:
            a_node = getattr(self, opr)(o_node)

        except AttributeError as e:

            if self.is_inline(o_node):
                a_node = self.cast_air(o_node.get("head"))

            else:
                # late catcher
                a_node = False

        except Exception as e:
            tb.print_exc()
            sys.exit(1)
        # if a_node: a_node = self.add_location(a_node, o_node)
        if a_node and level == self._top_level and not type(a_node) in self._stmt_types:
            a_node = Expr(a_node)
        return a_node

    def air_nop(self, *args, **kwds):
        # does nothing; can be use to disable nodes
        # TODO: handle in nested
        return None

    def air_say(self, data):
        # IR for print
        value = None
        p_args = []

        if self.is_inline(data):
            p_args = self._handle_inline_node(data.get("head"), "token")

        else:
            p_args = self._handle_full_node(data.get("body"))
        node = value = Call(func=Name(id="print", ctx=Load()), args=p_args, keywords=[])
        return node

    def air_setv(self, data):
        # assignment (simple for now)
        name = value = None
        targets = []

        if self.is_inline(data):
            data = self._behead(data.get("head"))
            name, value = data.split(" ", 1)
            name = Name(name, Store())
            value = self.cast_air(value)
            targets.append(name)

        elif self.is_full(data):
            anl = self._handle_full_node(data.get("body"))
            name = anl[0]["head"] if isinstance(anl[0], dict) else anl[0]
            name.ctx = Store()
            value = (
                self.cast_air(anl[-1]["head"]) if isinstance(anl[-1], dict) else anl[-1]
            )
            targets.append(name)
        node = Assign(targets=targets, value=value)
        return node

    def air_add(self, data):
        # addition
        return self._binop(data, "add")

    def air_sub(self, data):
        return self._binop(data, "sub")

    def air_mul(self, data):
        return self._binop(data, "mul")

    def air_div(self, data):
        return self._binop(data, "div")

    def _binop(self, data, opn):
        # arithmetic binary operation
        node = None
        bopl = bopr = None
        op = self._binop_map.get(opn.split("_", 1)[-1:][0])
        values = []

        if self.is_inline(data):
            data = self._behead(data.get("head")).split(" ")
            if not all([self.is_num(e) for e in data]):
                raise ValueError(
                    "All operands of a binary operation must be numbers [1]"
                )
            data = [self.cast_air(itm) for itm in data]

        else:
            data = self._handle_full_node(data.get("body"))
            data = [self._resolve(e) if self.is_astir(e) else e for e in data]
        opr_ts = [Num, BinOp]  # valid operand types
        if not all([type(e) in opr_ts for e in data]):
            raise ValueError("All operands of a binary operation must be numbers [2]")
        if len(data) == 1:
            raise ValueError("Only 1 number found during binary arithmetic operation")
        bopl = data[0]

        for itm in data[1:]:
            # recursive binary operations
            bopr = itm
            node = BinOp(left=bopl, op=op, right=bopr)
            bopl = node
        return node

    def is_astir(self, node):
        if not isinstance(node, dict):
            return False
        head = node.get("head")
        return True if head.startswith("air_") or self.cast_air(head, "test") else False

    def air_not(self, data):
        # unary not operation
        return self._unaryop(data, "not")

    def _unaryop(self, data, opn):
        node = None
        op = self._unaryop_map.get(opn.split("_", 1)[-1])

        if self.is_inline(data):
            data = self._behead(data.get("head"))
            data = self.cast_air(data)

        else:
            data = self._handle_full_node(data.get("body"))
            if not len(data) == 1:
                raise ValueError(
                    "A unary operation only takes 1 operand, but {len(data)} found"
                )
            data = data[0]
        node = UnaryOp(op, data)
        return node

    def air_and(self, data):
        # boolean and operation
        return self._boolop(data, "and")

    def air_or(self, data):
        return self._boolop(data, "or")

    def _boolop(self, data, opn):
        node = None
        op = self._boolop_map.get(opn.split("_", 1)[-1])
        values = []

        if self.is_inline(data):
            data = self._behead(data.get("head")).split(" ")
            data = [self.cast_air(v) for v in data]
            values = data

        else:
            data = self._handle_full_node(data.get("body"))
            values = data
        node = BoolOp(op, values)
        return node

    def air_eq(self, data):
        # equality comparison
        return self._compare(data, "eq")

    def air_not_eq(self, data):
        # equality comparison
        return self._compare(data, "not_eq")

    def air_lt(self, data):
        # equality comparison
        return self._compare(data, "lt")

    def air_lt_e(self, data):
        # equality comparison
        return self._compare(data, "lt_e")

    def air_gt(self, data):
        # equality comparison
        return self._compare(data, "gt")

    def air_gt_e(self, data):
        # equality comparison
        return self._compare(data, "gt_e")

    def _compare(self, data, opn):
        node = None
        op = self._compare_map.get(opn.split("_", 1)[-1])
        values = []

        if self.is_inline(data):
            data = self._handle_inline_node(data.get("head"), "split")
            values = data

        else:
            data = self._handle_full_node(data.get("body"))
            values = data
        if len(values) < 2:
            raise ValueError(
                f"Need at least 2 operands to do a comparison, but got {len(values)}"
            )
        left = values.pop(0)
        ops = [op] * len(values)
        node = Compare(left, ops, values)
        return node

    def air_call(self, data):
        # function call
        # TODO: handle eval'uable func name
        func = f_args = None

        if self.is_inline(data):
            data = self._behead(data.get("head"))
            if not " " in data:
                data += " "  # hack to handle no arg
            f_name, f_args = data.split(" ", 1)
            func = Name(id=f_name, ctx=Load())
            f_args = [self.cast_air(a) for a in self._tokenize(f_args)]

        else:
            data = self._handle_full_node(data.get("body"))
            if not data:
                raise ValueError("Need at least the name of a function to call")
            func = data.pop(0)
            f_args = data
        args = []
        kwds = []

        for a in f_args:
            if isinstance(a, keyword):
                kwds.append(a)

            else:
                args.append(a)
        node = Call(func=func, args=args, keywords=kwds)
        return node

    def air_import(self, data):
        # imports
        # TODO: support aliasing, multiple from and combined import methods
        node = None
        from_rx = re.compile(r"(?P<names>[\w\s*]+) from (?P<module>[\w.]+)")
        data = self._node_to_list(data)
        i_aliases = []
        fi_aliases = []

        for im in data:
            if isinstance(im, Name):
                # simple import
                i_aliases.append(alias(name=im.id, asname=None))
                continue
            match = from_rx.match(im.s)
            level = 0

            if match:
                # from import
                if len(data) > 1:
                    raise ImportError(
                        f"Currently only a single module can be imported from; got {len(data)} modules"
                    )
                match = match.groupdict()
                module = match.get("module")

                if re.match(r"^\.+$", module):
                    level = module.count(".")
                    module = None
                names = [
                    alias(name=a, asname=None) for a in match.get("names").split(" ")
                ]
                fi_aliases = {"module": module, "names": names, "level": level}
        if i_aliases and fi_aliases:
            raise ImportError(
                "Cannot currently combine both import methods under the same import node"
            )
        node = (
            Import(i_aliases)
            if i_aliases
            else ImportFrom(**fi_aliases)
            if fi_aliases
            else None
        )
        if not node:
            raise ImportError("Nothing found to import")
        return node

    def air_if(self, data):
        # condition
        return self._control_flow(data, "if")

    def _control_flow(self, data, cft, ignores=None, cl_cnt=None):
        if self.is_inline(data):
            raise Exception("Inline mode is not supported in control nodes")
        node = None
        clauses = []
        if not ignores:
            ignores = ["test", "body", "orelse"]
        if not cl_cnt:
            cl_cnt = [2, 3]  # clause count
        s_types = self._stmt_types
        cf = self._control_flow_map.get(cft)
        data = self._handle_full_node(data.get("body"), ignore_heads=ignores)
        if not len(data) in cl_cnt:
            raise ValueError(
                f"Expected {cl_cnt[0]} to {cl_cnt[1]} items, but got {len(data)}"
            )

        for idx, elt in enumerate(data):
            clause = elt if not isinstance(elt, dict) else elt.get(ignores[idx])

            if cft == "for":
                # for-specific element fixups
                if idx == 0 and isinstance(clause, Name):
                    clause.ctx = Store()

            if cft == "try" and idx == 1:
                # try-specific element fixups
                if not all([isinstance(e, ExceptHandler) for e in clause]):
                    raise ValueError(
                        'All nodes within the handlers clause must be of type "ExceptHandler"'
                    )

            if cft == "except" and idx == 1:
                if not isinstance(clause, Name):
                    raise NameError("Invalid name in except handler")
                clause = clause.id

            if idx + 1 in cl_cnt and not isinstance(clause, list):
                clause = [clause]
            clauses.append(clause)
        while True:
            # fill in remaining missing but required args
            if len(clauses) == cl_cnt[-1]:
                break
            clauses.append([])

        for clause in clauses:
            # wrap expressions to statements
            if not isinstance(clause, list):
                continue

            for idx, a_node in enumerate(clause):
                if not type(a_node) in s_types:
                    clause[idx] = Expr(a_node)
        node = cf(*clauses)
        return node

    def air_for(self, data):
        # loop
        ignores = ["target", "iter", "body", "orelse"]
        cl_cnt = [3, 4]
        return self._control_flow(data, "for", ignores, cl_cnt)

    def air_async_for(self, data):
        # loop
        ignores = ["target", "iter", "body", "orelse"]
        cl_cnt = [3, 4]
        return self._control_flow(data, "async_for", ignores, cl_cnt)

    def air_while(self, data):
        return self._control_flow(data, "while")

    def air_def(self, data):
        return self._function_def(data, FunctionDef)

    def air_async_def(self, data):
        return self._function_def(data, AsyncFunctionDef)

    def _function_def(self, data, nt):
        node = None
        o_clist = data.get("body")  # org clauses list
        f_name = o_clist[0].get("head")
        args = self._arguments(
            o_clist[1].get("body")
            if self.is_full(o_clist[1])
            else o_clist[1].get("head")
        )
        body = self._node_to_list(o_clist[2], ignore_first=["body"])
        body = [Expr(a_n) if not type(a_n) in self._stmt_types else a_n for a_n in body]
        decorators = []
        returns = None
        node = nt(f_name, args, body, decorators, returns)
        return node

    def _arguments(self, data):
        # process args
        node = None
        args_list = []
        vararg = None
        kwonlyargs = []
        kw_defaults = []
        kwarg = None
        defaults = []
        in_keys = False

        if isinstance(data, str):
            if data == "args":
                pass

            else:
                # inline args
                args = data.split(" ")
                if not all([self.is_name(a) for a in args]):
                    raise NameError("One or more invalid names found in arguments [1]")
                args = [arg(a, None) for a in args]
                args_list.append(args)

        elif isinstance(data, list):
            if any([" " in elt.get("head") for elt in data]):
                raise NameError("One or more invalid names found in arguments [2]")
            data = self._handle_full_node(data)

            for elt in data:

                if not in_keys:

                    if isinstance(elt, Name):
                        # pos arg, no def
                        a = arg(elt.id, None)
                        args_list.append(a)
                        continue

                    elif isinstance(elt, keyword):
                        # pos arg w/ def
                        a = arg(elt.arg, None)
                        d = elt.value
                        args_list.append(a)
                        defaults.append(d)
                        continue

                    elif isinstance(elt, Starred):
                        # starred arg (should contain Name node)
                        vararg = arg(elt.value.id, None)
                        in_keys = True
                        continue

                else:
                    if kwarg:
                        raise SyntaxError("Starred keyword must be the final argument")

                    if isinstance(elt, keyword):
                        # keyed arg or map
                        a = arg(elt.arg or "", None)
                        d = elt.value

                        if a.arg:
                            kwonlyargs.append(a)
                            kw_defaults.append(d)

                        else:
                            kwarg = arg(d.id, None) if isinstance(d, Name) else None
                            if not kwarg:
                                raise NameError(f"Expected a name-compatible str")

                    else:
                        raise SyntaxError(
                            "Cannot have a positional argument after a starred or keyword arg"
                        )
        node = arguments(args_list, vararg, kwonlyargs, kw_defaults, kwarg, defaults)
        return node

    def air_class(self, data):
        # class definition
        node = None
        o_clist = data.get("body")
        c_name = o_clist[0].get("head")
        bases = []
        keywords = []
        body = self._node_to_list(o_clist[1])
        body = [Expr(a_n) if not type(a_n) in self._stmt_types else a_n for a_n in body]
        decorators = []
        node = ClassDef(c_name, bases, keywords, body, decorators)
        return node

    def air_raise(self, data):
        # raise exceptions
        # TODO: handle cause clause
        return self._one_line_stmt(data, Raise, 1)

    def air_try(self, data):
        # monitor for and handle exceptions
        ignores = ["body", "handlers", "orelse", "finalbody"]
        cl_cnt = [2, 4]
        return self._control_flow(data, "try", ignores, cl_cnt)

    def air_except_handler(self, data):
        # exception handler (only valid within try handlers)
        ignores = ["type", "name", "body"]
        cl_cnt = [3, 3]
        return self._control_flow(data, "except", ignores, cl_cnt)

    def air_assert(self, data):
        data = self._node_to_list(data)
        if not len(data) == 2:
            raise ValueError(f"Expected 2 arguments but got {len(data)}")
        node = Assert(*data)
        return node

    def air_del(self, data):
        valid_nts = [Name, Attribute, Subscript]

        def fixup(node_list):
            for node in node_list:
                if hasattr(node, "ctx"):
                    node.ctx = Del()
            return node_list

        return self._one_line_stmt(data, Delete, "+", valid_nts, fixup)

    def air_pass(self, data):
        # placeholder
        return Pass()

    def air_break(self, data):
        # break out of a loop
        return Break()

    def air_continue(self, data):
        # continue a loop
        return Continue()

    def air_return(self, data):
        # return a value (at least None; cannot be empty)
        return self._one_line_stmt(data, Return, "1")

    def air_yield(self, data):
        #
        node = None
        return node

    def air_yield_from(self, data):
        #
        node = None
        return node

    def air_global(self, data):
        #
        node = None
        return node

    def air_nonlocal(self, data):
        #
        node = None
        return node

    def air_await(self, data):
        # used within async functions
        node = None
        return self._one_line_stmt(data, Await, "1")

    def _one_line_stmt(self, data, nt, arg_cnt="*", valid_nts=None, fixup=None):
        node = None
        value_list = self._node_to_list(data)
        if valid_nts and not all([type(elt) in valid_nts for elt in value_list]):
            raise ValueError("One or more invalid node types found")
        if str(arg_cnt).isdigit() and not len(value_list) == int(arg_cnt):
            raise ValueError(f"Expected {arg_cnt} nodes but got {len(value_list)}")
        if callable(fixup):
            value_list = fixup(value_list)

        if str(arg_cnt) == "1":
            node = nt(value_list[0])

        elif arg_cnt == "?":
            node = nt(value_list[0] if value_list else None)

        else:
            node = nt(value_list)
        return node

    def air_attr(self, data):
        # attribute access
        data = self._node_to_list(data)
        if not len(data) == 2:
            raise ValueError(f"Expected 2 parameters, but got {len(data)}")
        if not type(data[1]) in [Str, Name]:
            raise ValueError(f"Expected a string that can be used as a name")
        value = data[0]
        attr = data[1].s
        node = Attribute(value, attr, Load())
        return node

    def air_list(self, data):
        # list
        if not " " in data.get("head") and not data.get("body"):
            return List([], Load())
        data = self._node_to_list(data)
        node = List(data, Load())
        return node

    def air_tuple(self, data):
        # tuple
        if not " " in data.get("head") and not data.get("body"):
            return Tuple([], Load())
        data = self._node_to_list(data)
        node = Tuple(data, Load())
        return node

    def air_set(self, data):
        # set
        if not " " in data.get("head") and not data.get("body"):
            return Set([])
        data = self._node_to_list(data)
        node = Set(data)
        return node

    def air_dict(self, data):
        # dict
        keys = []
        values = []

        for pair in data.get("body"):
            key = self.cast_air(pair.get("head"))
            body = pair.get("body")
            value = (
                self._node_to_list(body[0])
                if len(body) == 1
                else self._handle_full_node(body)
                if len(body) > 1
                else None
            )
            if not value:
                raise ValueError("Missing corresponding value for key in dict")
            if isinstance(value, list):
                if not len(value) == 1:
                    raise ValueError(f"Expected a single value but got {len(value)}")
                value = value[0]
            keys.append(key)
            values.append(value)
        node = Dict(keys, values)
        return node

    def air_joined_str(self, data):
        # f-string
        data = self._node_to_list(data)
        values = [
            v if isinstance(v, Str) else FormattedValue(v, -1, None) for v in data
        ]
        node = JoinedStr(values)
        return node

    def air_starred(self, data):
        # variable reference
        node = self._one_line_stmt(data, Starred, "1")
        node.ctx = node.value.ctx
        return node

    def air_kwd(self, data):
        # keyword args
        node = None
        data = self._node_to_list(data)

        if len(data) == 1:
            # mapping ref
            node = keyword(None, data[0])

        elif len(data) == 2:
            if not isinstance(data[0], Name):
                raise NameError(f"Expected a name-compatible str")
            node = keyword(data[0].id, data[1])
        return node

    def _keyword(self, data):
        # Keyword arg in class def?
        arg = data.get("head")
        if not self.is_name(arg):
            raise NameError("Expecting a name-compatible str")
        body = data.get("body")
        value = self._node_to_list(body[0]) if len(body) == 1 else None
        if not value:
            raise ValueError(f"Expected a single valuue but got {len(body)}")
        node = keyword(arg, value)
        return node

    def air_slice(self, data):
        # subscripting simplified
        data = self._node_to_list(data)
        slice_ = None

        if len(data) == 2:
            slice_ = Index(data[1])

        elif len(data) in [3, 4]:
            slice_ = Slice(*(data[1:]))

        else:
            raise IndexError("Invalid number if subscripts found.")
        node = Subscript(data[0], slice_, Load())
        return node

    def air_with(self, data):
        return self._with_stmt(data, With)

    def air_async_with(self, data):
        return self._with_stmt(data, AsyncWith)

    def _with_stmt(self, data, nt):
        node = None
        ignores = ["items", "body"]
        s_types = self._stmt_types
        items = [self._withitem(itm) for itm in data["body"][0].get("body")]
        body = self._handle_full_node(data["body"][1].get("body"), ignore_heads=ignores)
        body = [Expr(b) if not type(b) in s_types else b for b in body]
        node = nt(items, body)
        return node

    def _withitem(self, data):
        node = None
        ctx_e = None
        opt_v = self._handle_inline_node(data.get("head"), "single")[0]

        if isinstance(opt_v, (Name, Tuple, List)):
            ctx_e = self._resolve(data.get("body")[0])
            if ctx_e in self._stmt_types + [Expr]:
                raise TypeError(
                    f'Item context for "with" must be an expression, not a statement'
                )
            if hasattr(opt_v, "ctx"):
                opt_v.ctx = Store()

        else:
            ctx_e = opt_v
            opt_v = None
        node = withitem(ctx_e, opt_v)
        return node


def generate_neu():
    """Make copy in neucode"""
    am = parse(open(sys.argv[0]).read())
    astir = "#! /usr/bin/env neu\n\n"

    for node in am.body[1:-2]:
        astir += f"* {dump(node)}\n"

    with open(sys.argv[0].replace(".py", ".neu", 1), "w") as af:
        af.write(astir)
    return


THIS_FILE = "astir.py"


if __name__ == "__main__":
    generate_neu()



"""
Python keywords made callable.
"""


from importlib import import_module


class CalleeException(Exception):
    pass


class BreakException(CalleeException):
    pass


class ContinueException(CalleeException):
    pass


class Callee:
    def __init__(self, *, namespace=None):
        if not isinstance(namespace, (dict, type(None))):
            raise ValueError('"namespace" must be a dict or None')
        self._vars = namespace or {}
        return

    @property
    def ns(self):
        return self._vars

    def x_setv(self, *names, value=NotImplemented):
        # NB: could blow up if someone actually wants to set NotImplemented
        names = list(names)
        if value is NotImplemented and len(names) > 1:
            value = names.pop(-1)
        if not all([isinstance(n, str) and n.isidentifier() for n in names]):
            raise NameError(f"names must be valid identifier(s)")
        if value is NotImplemented:
            raise ValueError("no value to set")
        for name in names:
            self._vars[name] = value
        return value

    def x_getv(self, *names):
        if not names:
            raise NameError("no name to get")
        if not all([isinstance(n, str) and n.isidentifier() for n in names]):
            raise NameError(f"names must be valid identifier(s)")
        values = []
        vars = self._vars

        for name in names:
            if name not in vars:
                raise NameError(f'name "{name}" is not defined')
            values.append(vars[name])
        return tuple(values) if len(values) > 1 else values[0]

    def x_delv(self, *names):
        if not names:
            raise NameError("no names to delete")
        if not all([isinstance(n, str) and n.isidentifier() for n in names]):
            raise ValueError("must be valid identifier(s)")
        values = []
        vars = self._vars

        for name in names:
            if not name in vars:
                raise NameError(f'name "{name}" is not defined')
            values.append(vars.pop(name))
        return values

    def x_if(self, test, body, orelse=None, **ctx):
        if not orelse:

            def orelse(**ctx):
                return retv

        if isinstance(test, str):

            def test(**ctx):
                return eval(test, globals(), ctx)

        if not all([callable(f) for f in [test, body, orelse]]):
            raise Exception("All must be callables")
        retv = None

        if test(**ctx):
            retv = body(**ctx)

        else:
            retv = orelse(**ctx)
        return retv

    def x_for(self, target, iter_, body, orelse=None, **ctx):
        if (
            not isinstance(target, str)
            or not target.isidentifier()
            and not all([s.isidentifier() for s in target.split(", ")])
        ):
            raise ValueError(
                '"target" must be a string identifier or comma separated identifiers'
            )
        iter(iter_)
        if orelse is None:

            def orelse(**ctx):
                return retv

        if not all([callable(f) for f in [body, orelse]]):
            raise ValueError("All must be callables")
        locs = {
            "iter_": iter_,
            "body": body,
            "orelse": orelse,
            "retv": None,
            "ctx": ctx,
        }

        f_loop = f"""
for {target} in iter_:
    try:
        retv = body({target}, **ctx)

    except BreakException:
        break

    except ContinueException:
        continue

else:
    retv = orelse(**ctx)"""
        exec(f_loop, globals(), locs)
        return locs.get("retv")

    def x_while(self, test, body, orelse=None, **ctx):
        if orelse is None:

            def orelse(**ctx):
                return retv

        if isinstance(test, str):
            test_str = test

            def test(**ctx):
                return eval(test_str, globals(), ctx)

        if not all([callable(f) for f in [test, body, orelse]]):
            raise Exception("All must be callables")
        retv = None

        while test(**ctx):
            try:
                retv = body(**ctx)

            except BreakException:
                break

            except ContinueException:
                continue

        else:
            retv = orelse(**ctx)
        return retv

    def x_break(self, *args, **kwds):
        raise BreakException()

    def x_continue(self, *args, **kwds):
        raise ContinueException()

    def x_do(self, *funcs, **ctx):
        if not funcs:
            return None
        funcs = [f for f in funcs]
        if isinstance(funcs[-1], dict) and not ctx:
            ctx = funcs.pop(-1)
        if not all([callable(f) for f in funcs]):
            raise ValueError("Must all be callables")
        retv = None

        for func in funcs:
            retv = func(**ctx)
            ctx["return"] = retv
        return retv

    def x_import(self, *names, packages=None):
        if not names:
            raise NameError("no names to import")
        if packages is None:
            packages = [None] * len(names)
        if isinstance(packages, str):
            packages = [packages]
        if not (
            isinstance(packages, (list, tuple))
            and len(packages) == len(names)
            and all([isinstance(p, (str, type(None))) for p in packages])
        ):
            raise ImportError(
                '"packages" must be a string or list of strings for packages corresponding to given names'
            )
        mods = []

        for name, package in zip(names, packages):
            mods.append(import_module(name, package))
        return tuple(mods) if len(mods) > 1 else mods[0]

    def x_with(self, *items, body=None, **ctx):
        if not items:
            raise ValueError("no items to create context")
        items = list(items)
        if body is None:
            body = items.pop(-1)
        if (
            len(items) == 2
            and callable(items[0])
            and isinstance(items[1], (str, type(None)))
        ):
            items = [items]
        if all(
            [
                (
                    isinstance(i, (list, tuple))
                    and len(i) == 2
                    and callable(i[0])
                    and isinstance(i[1], (str, type(None)))
                )
                for i in items
            ]
        ):
            items = {i[1]: i[0] for i in items}
        if not (
            isinstance(items, dict)
            and all([isinstance(n, (str, None)) for n in items.keys()])
            and all([callable(i) for i in items.values()])
        ):
            raise ValueError("Bad items format")
        if not callable(body):
            raise ValueError('"body" must be callable')
        header = ", ".join(
            [f"{v.__name__}()%s" % (f" as {k}" if k else "") for k, v in items.items()]
        )
        w_args = ", ".join([k for k in items if isinstance(k, str)])
        if w_args:
            w_args += ", "
        e_ctx = {c.__name__: c for c in items.values()}
        locs = {"body": body, "retv": None, "ctx": ctx, **e_ctx}
        w_ctx = f"""
with {header}:
    retv = body({w_args}**ctx)
        """
        exec(w_ctx, globals(), locs)
        return locs.get("retv")

    def x_class(self, name, attrdict, bases=None, decorators=None):
        retv = None
        if not (
            isinstance(attrdict, dict)
            and all([isinstance(k, str) and k.isidentifier() for k in attrdict])
        ):
            raise ValueError("Class attrdict must be a dict of attribs")
        if bases is None:
            bases = ()
        if decorators is None:
            decorators = []
        cls = type(name, bases, attrdict)

        for d in decorators:
            cls = d(cls)
        return cls

    def x_try(self, body, *handlers, orelse=None, finalbody=None, **ctx):
        if not callable(body):
            raise ValueError('"body" must be callable')
        if not (
            isinstance(handlers, (list, tuple))
            and all(
                [
                    (
                        isinstance(h, (list, tuple))
                        and len(h) == 2
                        and (
                            isinstance(h[0], Exception)
                            or hasattr(h[0], "with_traceback")
                        )
                        and callable(h[1])
                    )
                    for h in handlers
                ]
            )
        ):
            raise ValueError(
                '"handlers" must be a list or tuple of exception, callable handler lists/tuples'
            )
        if orelse is None:

            def orelse(**ctx):
                return retv

        if finalbody is None:

            def finalbody(**ctx):
                return retv

        if ctx is None:
            ctx = {}
        retv = None

        try:
            retv = body(**ctx)

        except Exception as exc:
            for h in handlers:
                if type(exc) == h[0]:
                    retv = h[1](exc, **ctx)
                    break

            else:
                raise

        else:
            retv = orelse(**ctx)

        finally:
            retv = finalbody(**ctx)
        return retv

    def x_raise(self, exc, *args, **kwds):
        if callable(exc) and isinstance(exc(), BaseException):
            exc = exc(*args, **kwds)
        if not isinstance(exc, BaseException):
            raise TypeError("exceptions must derive from BaseException")
        raise exc

    def to_dict(self):
        cs = {}

        for name in dir(self):
            if name.startswith("x_"):
                cs[name] = getattr(self, name)
        return cs



#! /usr/bin/env python

# This file is part of neulang

# neulang - A language sitting on top of Python that executes pseudocode that's very close to natural language.

# @author Andrew Phillips
# @copyright 2017-2019 Andrew Phillips <skeledrew@gmail.com>

# neulang is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.

# neulang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.

# You should have received a copy of the GNU Affero General Public
# License along with neulang.  If not, see <http://www.gnu.org/licenses/>.

import sys
from os.path import join, dirname

from setuptools import setup

from neulang.main import __meta__, __version__

setup(
    name=__meta__.get('name', 'neulang'),
    version=__version__,
    entry_points={
        'console_scripts': [
            'neu = neulang.main:run',
        ],
    },
    author='Andrew Phillips',
    author_email='skeledrew@gmail.com',
    description='Executable org-formatted pseudocode embedded in Python.',
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    long_description_content_type="text/markdown",
    url='https://gitlab.com/skeledrew/neulang',
    packages=['neulang', 'tests'],
    platforms=['any'],
    install_requires = [
        'docopt',
        'adapt-parser',
    ],
    license='AGPLv3',
    classifiers=[
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development',
    ])



"""

"""


import traceback as tb

import sys
import re
from ast import *

from .base import BaseResolver
from .astir import ASTIR
from .adapt_resolver import AdaptResolver


class Resolver(BaseResolver):
    def __init__(self, *, parent=None, pubsub=None):
        self._parent = parent
        super().__init__()
        del self._resolvers["Resolver"]
        PyExprResolver()
        ASTIR()
        # RegexResolver()
        AdaptResolver()
        assert pubsub and hasattr(pubsub, "pub"), "PubSub instance not available"

        for r in self.get_resolvers().values():
            # hack items into resolvers
            setattr(r, "_pubsub", pubsub)
        return

    def resolve(self, tree):
        """Top level delegating resolver"""
        if not tree:
            raise ValueError("Unable to find any parseable content. Missing *'s?")
        nodes = []
        self._top_level = tree[0].get("level")

        for node in tree:

            try:
                a_node = self._resolve(node)
                if a_node:
                    nodes.append(a_node) if not isinstance(
                        a_node, list
                    ) else nodes.extend(a_node)

            except Exception as e:
                tb.print_exc()
                return None
        assert not any([isinstance(n, list) for n in nodes])
        module = Module(nodes)
        return module

    def get_pif(self, key):
        resolvers = list(self._resolvers.values())
        pif = None

        for r in resolvers:
            pif = getattr(r, "_pifs", {}).get(key)
            if pif:
                break
        if not pif:
            raise AttributeError(f'Unable to resolve PIF "{key}"')
        return pif

    def add_handler(self, data):
        return None


class RegexResolver(BaseResolver):
    def __init__(self):
        super().__init__()
        self._rx_test = re.compile(r"(^add_regex_handler(\s.+)?)")
        self._rx_hashes = {}
        self._pifs = {}

    def resolve(self, o_node):
        if not self.is_regex(o_node):
            return False
        a_node = None
        head = o_node.get("head")
        h_match = re.match(r"add_(?P<type>\w+)_handler(\s.+)?", head)

        if h_match:
            # TODO: handler type detection in 'self.add_handler'
            h_type = h_match.groupdict().get("type")
            a_node = self.add_handler(o_node)
            if a_node is not None:
                raise ValueError(f"None expected but got {type(a_node)}")

        else:
            a_node = self.nl_call(o_node)
        return a_node

    def add_handler(self, data):
        ignores = ["match", "body"]
        data = self._node_to_list(data, ignore_first=ignores)
        if not len(data) == 2:
            raise ValueError(f"Expected 2")
        match = data[0].get("match")[0]
        if not isinstance(match, Str):
            raise ValueError(f"Expected string")
        match = match.s
        body = data[1].get("body")
        body = [Expr(a_n) if not type(a_n) in self._stmt_types else a_n for a_n in body]
        rx_h = f"rx_{abs(hash(match))}"
        self._rx_hashes[match] = rx_h
        pif = body  # pif as a snippet
        self._pifs[rx_h] = pif
        return None

    def nl_call(self, data):
        head = data.get("head")
        body = data.get("body", [])
        rx_h = context = p_body = None

        if body:
            # preprocess body (allows for late handler definitions)
            p_body = self._node_to_list(data)

        for rx in self._rx_hashes:
            # find an appropriate handler
            # TODO: make abstract for any handler type
            match = re.match(rf"{rx}", head)
            if not match:
                continue
            rx_h = self._rx_hashes.get(rx)
            context = match.groupdict()
            context = parse(f"context={context}").body[0]  # get dict node
            break
        if not rx_h:
            return False  # late detection of handlability
        pif = self._pifs[rx_h]
        assert isinstance(pif, list), f"Expected a list but got (type(pif))"
        node = [context] + pif
        node.extend(p_body or [])
        return node

    def is_regex(self, node):
        # TODO: better way to detect handlable nodes
        if not isinstance(node, dict):
            return False
        head = node.get("head", "")
        return (
            True
        )  # if self._rx_test.match(head) or not self.cast_air(head, 'test') else False


class PyExprResolver(BaseResolver):
    # TODO: maybe move to own module

    def __init__(self):
        super().__init__()
        return

    def resolve(self, data):
        """Convert an expression to a node"""
        if not self.is_expr(data):
            return False
        # TODO: ensure inline? method to incl body node in call?
        head = data.get("head")
        if compilable(head):
            node = parse(head).body[0].value

        else:
            # assumes this is the last to be resolved
            node = Str(head)
        return node

    def add_handler(self, data):
        # actually shouldn't be necessary in this class
        return None

    def is_expr(self, node):
        """Return True if a node is a Python expression."""
        if not isinstance(node, dict):
            return False
        text = node.get("head")

        try:
            # if re.match(r'[\w.]+\(.*\)', text): return True
            if " " in text:
                return True
            compile(text, "<expr_test>", "eval")
            return True

        except SyntaxError:
            pass
        return False


def compilable(text, mode="exec"):
    """Return True if a string can be compiled"""
    # TODO: move to utils
    if not mode in ["eval", "exec"]:
        raise ValueError(f'"{mode}" is an invalid compile mode')

    try:
        compile(text, "", mode)
        return True

    except:
        pass
    return False



"""

"""


class Life:
    PERMANENT = -1


class PubSub:
    def __init__(self):
        self._hub = hub = {}
        return

    def pub(self, topic, content, *, life=Life.PERMANENT):
        hub = self._hub
        if topic not in hub:
            hub[topic] = {"content_list": [], "life": life, "callbacks": []}

        if hub[topic]["life"] is None:
            hub[topic]["life"] = life
        hub[topic]["life"] == Life.PERMANENT
        hub[topic]["content"] = content

        if hub[topic]["life"] == Life.PERMANENT:
            hub[topic]["content_list"].append(content)

        for c in hub[topic]["callbacks"]:
            c(content)
        return True

    def sub(self, *, topic=None, callback=None):
        hub = self._hub

        if not (topic or callback):
            return list(hub.keys())

        if not (topic and isinstance(topic, str)):
            raise TypeError('"topic" must be a string')

        if callback is None:
            matches = []

            for t in hub:
                if t.startswith(topic):
                    matches.append(t)
            return matches

        else:
            if not callable(callback):
                raise ValueError('"callback" must be a callable')
            if topic not in hub:
                hub[topic] = {"callbacks": [], "life": None, "content_list": []}
            hub[topic]["callbacks"].append(callback)

            if hub[topic]["life"] in [Life.PERMANENT]:

                for c in hub[topic]["content_list"]:
                    callback(c)
        return True
